/**
 * Copyright (c) 2025 Enzien Audio, Ltd.
 * 
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions, and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the phrase "powered by heavy",
 *    the heavy logo, and a hyperlink to https://enzienaudio.com, all in a visible
 *    form.
 * 
 *   2.1 If the Application is distributed in a store system (for example,
 *       the Apple "App Store" or "Google Play"), the phrase "powered by heavy"
 *       shall be included in the app description or the copyright text as well as
 *       the in the app itself. The heavy logo will shall be visible in the app
 *       itself as well.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */

#include "Heavy_ChaffVerb.h"
#include "HeavyDPF_ChaffVerb.hpp"
#include <set>


#define HV_LV2_NUM_PARAMETERS 23

#define HV_HASH_NOTEIN          0x67E37CA3
#define HV_HASH_CTLIN           0x41BE0f9C
#define HV_HASH_POLYTOUCHIN     0xBC530F59
#define HV_HASH_PGMIN           0x2E1EA03D
#define HV_HASH_TOUCHIN         0x553925BD
#define HV_HASH_BENDIN          0x3083F0F7
#define HV_HASH_MIDIIN          0x149631bE
#define HV_HASH_MIDIREALTIMEIN  0x6FFF0BCF

#define HV_HASH_NOTEOUT         0xD1D4AC2
#define HV_HASH_CTLOUT          0xE5e2A040
#define HV_HASH_POLYTOUCHOUT    0xD5ACA9D1
#define HV_HASH_PGMOUT          0x8753E39E
#define HV_HASH_TOUCHOUT        0x476D4387
#define HV_HASH_BENDOUT         0xE8458013
#define HV_HASH_MIDIOUT         0x6511DE55
#define HV_HASH_MIDIOUTPORT     0x165707E4

#define MIDI_RT_CLOCK           0xF8
#define MIDI_RT_START           0xFA
#define MIDI_RT_CONTINUE        0xFB
#define MIDI_RT_STOP            0xFC
#define MIDI_RT_ACTIVESENSE     0xFE
#define MIDI_RT_RESET           0xFF

#define HV_HASH_DPF_BPM         0xDF8C2721

// midi realtime messages
std::set<int> mrtSet {
  MIDI_RT_CLOCK,
  MIDI_RT_START,
  MIDI_RT_CONTINUE,
  MIDI_RT_STOP,
  MIDI_RT_RESET
};


START_NAMESPACE_DISTRHO


// -------------------------------------------------------------------
// Heavy Send and Print hooks

static void hvSendHookFunc(HeavyContextInterface *c, const char *sendName, uint32_t sendHash, const HvMessage *m)
{
  HeavyDPF_ChaffVerb* plugin = (HeavyDPF_ChaffVerb*)c->getUserData();
  if (plugin != nullptr)
  {
  }
}

static void hvPrintHookFunc(HeavyContextInterface *c, const char *printLabel, const char *msgString, const HvMessage *m)
{
  char buf[64];
  char* dst = buf;
  int len = strnlen(printLabel, 48);
  dst = strncpy(dst, printLabel, len);
  dst = strcpy(dst, " ");
  dst = strncpy(dst, msgString, 63-len);
  printf("> %s \n", buf);
}

// -------------------------------------------------------------------
// Main DPF plugin class

HeavyDPF_ChaffVerb::HeavyDPF_ChaffVerb()
 : Plugin(HV_LV2_NUM_PARAMETERS, 0, 0)
{
  _parameters[0] = 1500.0f;
  _parameters[1] = 1.0f;
  _parameters[2] = 4.0f;
  _parameters[3] = 0.0f;
  _parameters[4] = 0.0f;
  _parameters[5] = 0.0f;
  _parameters[6] = 1.0f;
  _parameters[7] = 0.0f;
  _parameters[8] = 1500.0f;
  _parameters[9] = 0.0f;
  _parameters[10] = 0.0f;
  _parameters[11] = 0.0f;
  _parameters[12] = 0.0f;
  _parameters[13] = 0.7f;
  _parameters[14] = 800.0f;
  _parameters[15] = 0.0f;
  _parameters[16] = 1000.0f;
  _parameters[17] = 500.0f;
  _parameters[18] = 0.65f;
  _parameters[19] = 0.45f;
  _parameters[20] = 0.1f;
  _parameters[21] = 0.85f;
  _parameters[22] = 0.5f;
  

  _context = hv_ChaffVerb_new_with_options(getSampleRate(), 10, 23, 2);
  _context->setUserData(this);
  _context->setSendHook(&hvSendHookFunc);
  _context->setPrintHook(&hvPrintHookFunc);

  
  // ensure that the new context has the current parameters
  for (int i = 0; i < HV_LV2_NUM_PARAMETERS; ++i) {
    setParameterValue(i, _parameters[i]);
  }
}

HeavyDPF_ChaffVerb::~HeavyDPF_ChaffVerb() {
  hv_ChaffVerb_free(_context);
}

void HeavyDPF_ChaffVerb::initParameter(uint32_t index, Parameter& parameter)
{// initialise parameters with defaults
  switch (index)
  {
    case paramA__LFO_Period:
        parameter.name = "A  LFO Period";
        parameter.symbol = "a__lfo_period";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = 1.0f;
        parameter.ranges.max = 4000.0f;
        parameter.ranges.def = 1500.0f;
        break;
    case paramA__LFO_Steps_Scale__Type2:
        parameter.name = "A  LFO Steps Scale  Type2";
        parameter.symbol = "a__lfo_steps_scale__type2";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = -5.0f;
        parameter.ranges.max = 5.0f;
        parameter.ranges.def = 1.0f;
        break;
    case paramA__LFO_Steps__Type2:
        parameter.name = "A  LFO Steps  Type2";
        parameter.symbol = "a__lfo_steps__type2";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = 2.0f;
        parameter.ranges.max = 10.0f;
        parameter.ranges.def = 4.0f;
        break;
    case paramA__LFO_Switch:
        parameter.name = "A  LFO Switch";
        parameter.symbol = "a__lfo_switch";
        parameter.hints = kParameterIsAutomatable
        | kParameterIsBoolean;
        parameter.ranges.min = 0.0f;
        parameter.ranges.max = 1.0f;
        parameter.ranges.def = 0.0f;
        break;
    case paramA__LFO_type:
        parameter.name = "A  LFO type";
        parameter.symbol = "a__lfo_type";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = 0.0f;
        parameter.ranges.max = 2.99f;
        parameter.ranges.def = 0.0f;
        break;
    case paramB__General_Offset:
        parameter.name = "B  General Offset";
        parameter.symbol = "b__general_offset";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = -2.0f;
        parameter.ranges.max = 2.0f;
        parameter.ranges.def = 0.0f;
        break;
    case paramB__General_Scale:
        parameter.name = "B  General Scale";
        parameter.symbol = "b__general_scale";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = -7.0f;
        parameter.ranges.max = 7.0f;
        parameter.ranges.def = 1.0f;
        break;
    case paramC__Random_LFO_Offset:
        parameter.name = "C  Random LFO Offset";
        parameter.symbol = "c__random_lfo_offset";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = -2.0f;
        parameter.ranges.max = 2.0f;
        parameter.ranges.def = 0.0f;
        break;
    case paramC__Random_LFO_Period:
        parameter.name = "C  Random LFO Period";
        parameter.symbol = "c__random_lfo_period";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = 1.0f;
        parameter.ranges.max = 4000.0f;
        parameter.ranges.def = 1500.0f;
        break;
    case paramC__Random_LFO_Scale:
        parameter.name = "C  Random LFO Scale";
        parameter.symbol = "c__random_lfo_scale";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = -7.0f;
        parameter.ranges.max = 7.0f;
        parameter.ranges.def = 0.0f;
        break;
    case paramC__Random_LFO_Switch:
        parameter.name = "C  Random LFO Switch";
        parameter.symbol = "c__random_lfo_switch";
        parameter.hints = kParameterIsAutomatable
        | kParameterIsBoolean;
        parameter.ranges.min = 0.0f;
        parameter.ranges.max = 1.0f;
        parameter.ranges.def = 0.0f;
        break;
    case paramD__Shift_Microtone:
        parameter.name = "D  Shift Microtone";
        parameter.symbol = "d__shift_microtone";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = -1.0f;
        parameter.ranges.max = 1.0f;
        parameter.ranges.def = 0.0f;
        break;
    case paramD__Shift_Transpose:
        parameter.name = "D  Shift Transpose";
        parameter.symbol = "d__shift_transpose";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = -36.0f;
        parameter.ranges.max = 36.0f;
        parameter.ranges.def = 0.0f;
        break;
    case paramE__Fx_Level:
        parameter.name = "E  Fx Level";
        parameter.symbol = "e__fx_level";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = 0.0f;
        parameter.ranges.max = 1.0f;
        parameter.ranges.def = 0.7f;
        break;
    case paramE__HiPass_Freq:
        parameter.name = "E  HiPass Freq";
        parameter.symbol = "e__hipass_freq";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = 100.0f;
        parameter.ranges.max = 2500.0f;
        parameter.ranges.def = 800.0f;
        break;
    case paramE__Shift_Delay:
        parameter.name = "E  Shift Delay";
        parameter.symbol = "e__shift_delay";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = 0.0f;
        parameter.ranges.max = 500.0f;
        parameter.ranges.def = 0.0f;
        break;
    case paramE__Shift_Window:
        parameter.name = "E  Shift Window";
        parameter.symbol = "e__shift_window";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = 1.0f;
        parameter.ranges.max = 3500.0f;
        parameter.ranges.def = 1000.0f;
        break;
    case paramF__FX_Delay:
        parameter.name = "F  FX Delay";
        parameter.symbol = "f__fx_delay";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = 0.0f;
        parameter.ranges.max = 2000.0f;
        parameter.ranges.def = 500.0f;
        break;
    case paramF__FX_Delay_Feedback:
        parameter.name = "F  FX Delay Feedback";
        parameter.symbol = "f__fx_delay_feedback";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = 0.0f;
        parameter.ranges.max = 0.97f;
        parameter.ranges.def = 0.65f;
        break;
    case paramMix__Dry_Mix:
        parameter.name = "Mix  Dry Mix";
        parameter.symbol = "mix__dry_mix";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = 0.0f;
        parameter.ranges.max = 1.0f;
        parameter.ranges.def = 0.45f;
        break;
    case paramMix__Dry_Vol:
        parameter.name = "Mix  Dry Vol";
        parameter.symbol = "mix__dry_vol";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = 0.0f;
        parameter.ranges.max = 1.0f;
        parameter.ranges.def = 0.1f;
        break;
    case paramMix__Reverb_Level:
        parameter.name = "Mix  Reverb Level";
        parameter.symbol = "mix__reverb_level";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = 0.0f;
        parameter.ranges.max = 1.0f;
        parameter.ranges.def = 0.85f;
        break;
    case paramMix__Reverb_Vol:
        parameter.name = "Mix  Reverb Vol";
        parameter.symbol = "mix__reverb_vol";
        parameter.hints = kParameterIsAutomatable;
        parameter.ranges.min = 0.0f;
        parameter.ranges.max = 1.0f;
        parameter.ranges.def = 0.5f;
        break;
    
  }
  
}

// -------------------------------------------------------------------
// Internal data

float HeavyDPF_ChaffVerb::getParameterValue(uint32_t index) const
{
  return _parameters[index];
  
}

void HeavyDPF_ChaffVerb::setParameterValue(uint32_t index, float value)
{
  switch (index) {
    case 0: {
      _context->sendFloatToReceiver(
          Heavy_ChaffVerb::Parameter::In::A__LFO_PERIOD,
          value);
      break;
    }
    case 1: {
      _context->sendFloatToReceiver(
          Heavy_ChaffVerb::Parameter::In::A__LFO_STEPS_SCALE__TYPE2,
          value);
      break;
    }
    case 2: {
      _context->sendFloatToReceiver(
          Heavy_ChaffVerb::Parameter::In::A__LFO_STEPS__TYPE2,
          value);
      break;
    }
    case 3: {
      _context->sendFloatToReceiver(
          Heavy_ChaffVerb::Parameter::In::A__LFO_SWITCH,
          value);
      break;
    }
    case 4: {
      _context->sendFloatToReceiver(
          Heavy_ChaffVerb::Parameter::In::A__LFO_TYPE,
          value);
      break;
    }
    case 5: {
      _context->sendFloatToReceiver(
          Heavy_ChaffVerb::Parameter::In::B__GENERAL_OFFSET,
          value);
      break;
    }
    case 6: {
      _context->sendFloatToReceiver(
          Heavy_ChaffVerb::Parameter::In::B__GENERAL_SCALE,
          value);
      break;
    }
    case 7: {
      _context->sendFloatToReceiver(
          Heavy_ChaffVerb::Parameter::In::C__RANDOM_LFO_OFFSET,
          value);
      break;
    }
    case 8: {
      _context->sendFloatToReceiver(
          Heavy_ChaffVerb::Parameter::In::C__RANDOM_LFO_PERIOD,
          value);
      break;
    }
    case 9: {
      _context->sendFloatToReceiver(
          Heavy_ChaffVerb::Parameter::In::C__RANDOM_LFO_SCALE,
          value);
      break;
    }
    case 10: {
      _context->sendFloatToReceiver(
          Heavy_ChaffVerb::Parameter::In::C__RANDOM_LFO_SWITCH,
          value);
      break;
    }
    case 11: {
      _context->sendFloatToReceiver(
          Heavy_ChaffVerb::Parameter::In::D__SHIFT_MICROTONE,
          value);
      break;
    }
    case 12: {
      _context->sendFloatToReceiver(
          Heavy_ChaffVerb::Parameter::In::D__SHIFT_TRANSPOSE,
          value);
      break;
    }
    case 13: {
      _context->sendFloatToReceiver(
          Heavy_ChaffVerb::Parameter::In::E__FX_LEVEL,
          value);
      break;
    }
    case 14: {
      _context->sendFloatToReceiver(
          Heavy_ChaffVerb::Parameter::In::E__HIPASS_FREQ,
          value);
      break;
    }
    case 15: {
      _context->sendFloatToReceiver(
          Heavy_ChaffVerb::Parameter::In::E__SHIFT_DELAY,
          value);
      break;
    }
    case 16: {
      _context->sendFloatToReceiver(
          Heavy_ChaffVerb::Parameter::In::E__SHIFT_WINDOW,
          value);
      break;
    }
    case 17: {
      _context->sendFloatToReceiver(
          Heavy_ChaffVerb::Parameter::In::F__FX_DELAY,
          value);
      break;
    }
    case 18: {
      _context->sendFloatToReceiver(
          Heavy_ChaffVerb::Parameter::In::F__FX_DELAY_FEEDBACK,
          value);
      break;
    }
    case 19: {
      _context->sendFloatToReceiver(
          Heavy_ChaffVerb::Parameter::In::MIX__DRY_MIX,
          value);
      break;
    }
    case 20: {
      _context->sendFloatToReceiver(
          Heavy_ChaffVerb::Parameter::In::MIX__DRY_VOL,
          value);
      break;
    }
    case 21: {
      _context->sendFloatToReceiver(
          Heavy_ChaffVerb::Parameter::In::MIX__REVERB_LEVEL,
          value);
      break;
    }
    case 22: {
      _context->sendFloatToReceiver(
          Heavy_ChaffVerb::Parameter::In::MIX__REVERB_VOL,
          value);
      break;
    }
    
    default: return;
  }
  _parameters[index] = value;
  
}


// -------------------------------------------------------------------
// Process

// void HeavyDPF_ChaffVerb::activate()
// {

// }

// void HeavyDPF_ChaffVerb::deactivate()
// {

// }



// -------------------------------------------------------------------
// DPF Plugin run() loop

#if DISTRHO_PLUGIN_WANT_MIDI_INPUT
void HeavyDPF_ChaffVerb::run(const float** inputs, float** outputs, uint32_t frames, const MidiEvent* midiEvents, uint32_t midiEventCount)
{
  handleMidiInput(frames, midiEvents, midiEventCount);
#else
void HeavyDPF_ChaffVerb::run(const float** inputs, float** outputs, uint32_t frames)
{
#endif
  const TimePosition& timePos(getTimePosition());
  if (timePos.playing && timePos.bbt.valid)
    _context->sendMessageToReceiverV(HV_HASH_DPF_BPM, 0, "f", timePos.bbt.beatsPerMinute);

  _context->process((float**)inputs, outputs, frames);
}

// -------------------------------------------------------------------
// Callbacks

void HeavyDPF_ChaffVerb::sampleRateChanged(double newSampleRate)
{
  hv_ChaffVerb_free(_context);

  _context = hv_ChaffVerb_new_with_options(getSampleRate(), 10, 23, 2);
  _context->setUserData(this);
  _context->setSendHook(&hvSendHookFunc);
  _context->setPrintHook(&hvPrintHookFunc);

  // ensure that the new context has the current parameters
  for (int i = 0; i < HV_LV2_NUM_PARAMETERS; ++i) {
    setParameterValue(i, _parameters[i]);
  }
}


// -----------------------------------------------------------------------
/* Plugin entry point, called by DPF to create a new plugin instance. */

Plugin* createPlugin()
{
    return new HeavyDPF_ChaffVerb();
}

// -----------------------------------------------------------------------

END_NAMESPACE_DISTRHO